/*---------------------------------------------------------------------------*/

/*!  \file  blend.h
     \brief  definitions of the polymer blends.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2007 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#ifndef INCLUDE_BLEND_H
#define INCLUDE_BLEND_H

#include "geometry.h"

/*---------------------------------------------------------------------------*/
/*!  \brief  polymer blend  */
typedef struct polymer_blend
{
        /*!  \brief  number of particle species.  */
    int n;
        /*!  \brief  numbers of soft particles.  */
    int *m;

        /*!  \brief  index of connected next block (negative if not connected).  */
    int *next;

        /*!  \brief  interaction parameter.  */
    double **epsilon;

        /*!  \brief  square root of 2 dimensional total particle density.  */
    double lambda0;
} polymer_blend;

/*---------------------------------------------------------------------------*/
/*!  \brief  allocate a new polymer blend
     \param  n  number of polymer species  */
polymer_blend *allocate_polymer_blend(int n);
/*!  \brief  free a polymer blend
     \param  blend  polymer blend to free  */
void free_polymer_blend(polymer_blend *blend);

/*!  \brief  calculate the square root of 2 dimensional total particle density.
     \param  blend  polymer blend
     \param  geometry  simulation box geometry  */
void calculate_lambda0(polymer_blend *blend,box_geometry *geometry);

/*---------------------------------------------------------------------------*/
#endif

/*---------------------------------------------------------------------------*/
