dnl ---------------------------------------------------------------------------
dnl
dnl  force.c.m4
dnl
dnl ---------------------------------------------------------------------------
dnl
dnl  Copyright (C) 2007,2008 Takashi Uneyama
dnl
dnl  This program is free software; you can redistribute it and/or modify
dnl  it under the terms of the GNU General Public License as published by
dnl  the Free Software Foundation; either version 2, or (at your option)
dnl  any later version.
dnl
dnl  This program is distributed in the hope that it will be useful,
dnl  but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl  GNU General Public License for more details.
dnl
dnl  You should have received a copy of the GNU General Public License
dnl  along with this program; if not, write to the Free Software
dnl  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
dnl  USA.
dnl
dnl ---------------------------------------------------------------------------
include(`dimension.m4')`'dnl
GENERATE_HEADER(force.c,
                calculate the random and interaction forces.)

#include "common.h"

#ifdef HAVE_MATH_H
#include <math.h>
#endif

#include "vector.h"
#include "particles.h"
#include "nrand.h"
#include "force.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif
#ifndef M_SQRTPI
#define M_SQRTPI 1.77245385090551602729
#endif

#define SQR(x) ((x)*(x))

#define INDEX(i,n) (((i) == -1) ? ((n) - 1) : (((i) == (n)) ? (0) : (i)))
#define LATTICE_VECTOR(i,n,c) (((i) == -1) ? (c) : (((i) == (n)) ? (-(c)) : (0)))
#define IS_LOWER_BOUNDARY(i1,i2,n) (((i1) == 0) && ((i2) == -1))
#define IS_UPPER_BOUNDARY(i1,i2,n) (((i1) == ((n) - 1)) && ((i2) == (n)))

/*---------------------------------------------------------------------------*/
PROTOTYPE_FUNCTIONS(`dnl
static void generate_random_force@D@(soft_particles@D@ *particles);
static void clear_interaction_force@D@(soft_particles@D@ *particles);
static void clear_stress_tensor@D@(soft_particles@D@ *particles);'`')

static void accumulate_tethering_interaction_stress(soft_particles *particles,
                                                    polymer_blend *blend,
                                                    box_geometry *geometry);
static void accumulate_site_interaction_stress(soft_particles *particles,
                                               linked_cell_list *celllist,
                                               polymer_blend *blend,
                                               box_geometry *geometry);

dnl ---------------------------------------------------------------------------
dnl  GENERATE_FORCE_CALCULATION_PROTOTYPE
dnl    generate prototypes for the force or the stress tensor calculations.
dnl  $1 the function mode (force or stress).
dnl  $2 the target argument (vector@D@ or symmetric_tensor@D@).
dnl  $3 the name of the target argument (f or sigma).
dnl ---------------------------------------------------------------------------
define(`GENERATE_FORCE_CALCULATION_PROTOTYPE',`dnl
PROTOTYPE_FUNCTIONS(`dnl
static void accumulate_tethering_interaction_$1@D@(vector@D@ **r,$2 **$3,
                                                    polymer_blend *blend,
                                                    box_geometry *geometry);
static void accumulate_site_interaction_$1@D@(vector@D@ **r,$2 **$3,
                                               linked_cell_list *celllist,
                                               polymer_blend *blend,
                                               box_geometry *geometry);
static void accumulate_site_interaction_$1@D@_in_different_cells(
    vector@D@ **r,$2 **$3,
    linked_cell_list *celllist,polymer_blend *blend,
    int k1,int k2,vector@D@ *c);
static void accumulate_site_interaction_$1@D@_in_same_cell(
    vector@D@ **r,$2 **$3,
    linked_cell_list *celllist,polymer_blend *blend,
    int k,vector@D@ *c);'`')'`')
dnl ---------------------------------------------------------------------------
dnl
GENERATE_FORCE_CALCULATION_PROTOTYPE(force,vector@D@,f)

GENERATE_FORCE_CALCULATION_PROTOTYPE(stress,symmetric_tensor@D@,sigma)

/*---------------------------------------------------------------------------*/
void generate_random_force(soft_particles *particles)
{
    SWITCH_FUNCTION(particles->dimension,
                    generate_random_force@D@(&(particles->particles@D@));)
}

/*---------------------------------------------------------------------------*/
void clear_interaction_force(soft_particles *particles)
{
    SWITCH_FUNCTION(particles->dimension,
                    clear_interaction_force@D@(&(particles->particles@D@));)
}

/*---------------------------------------------------------------------------*/
void clear_stress_tensor(soft_particles *particles)
{
    SWITCH_FUNCTION(particles->dimension,
                    clear_stress_tensor@D@(&(particles->particles@D@));)
}

/*---------------------------------------------------------------------------*/
void accumulate_tethering_interaction_force(soft_particles *particles,
                                            polymer_blend *blend,
                                            box_geometry *geometry)
{
    SWITCH_FUNCTION(particles->dimension,
                    accumulate_tethering_interaction_force@D@(particles->particles@D@.r,
                        particles->particles@D@.f,blend,geometry);)
}

/*---------------------------------------------------------------------------*/
void accumulate_tethering_interaction_force_star(soft_particles *particles,
                                                 polymer_blend *blend,
                                                 box_geometry *geometry)
{
    SWITCH_FUNCTION(particles->dimension,
                    accumulate_tethering_interaction_force@D@(particles->particles@D@.r_star,
                        particles->particles@D@.f_star,blend,geometry);)
}

/*---------------------------------------------------------------------------*/
void accumulate_site_interaction_force(soft_particles *particles,
                                       linked_cell_list *celllist,
                                       polymer_blend *blend,
                                       box_geometry *geometry)
{
    SWITCH_FUNCTION(particles->dimension,
                    accumulate_site_interaction_force@D@(particles->particles@D@.r,
                                                       particles->particles@D@.f,
                                                       celllist,blend,geometry);)
}

/*---------------------------------------------------------------------------*/
void accumulate_site_interaction_force_star(soft_particles *particles,
                                            linked_cell_list *celllist,
                                            polymer_blend *blend,
                                            box_geometry *geometry)
{
    SWITCH_FUNCTION(particles->dimension,
                    accumulate_site_interaction_force@D@(particles->particles@D@.r_star,
                                                       particles->particles@D@.f_star,
                                                       celllist,blend,geometry);)
}

/*---------------------------------------------------------------------------*/
void accumulate_tethering_interaction_stress(soft_particles *particles,
                                            polymer_blend *blend,
                                            box_geometry *geometry)
{
    SWITCH_FUNCTION(particles->dimension,
                    accumulate_tethering_interaction_stress@D@(particles->particles@D@.r,
                        particles->particles@D@.sigma,blend,geometry);)
}

/*---------------------------------------------------------------------------*/
void accumulate_site_interaction_stress(soft_particles *particles,
                                       linked_cell_list *celllist,
                                       polymer_blend *blend,
                                       box_geometry *geometry)
{
    SWITCH_FUNCTION(particles->dimension,
                    accumulate_site_interaction_stress@D@(particles->particles@D@.r,
                                                       particles->particles@D@.sigma,
                                                       celllist,blend,geometry);)
}

/*---------------------------------------------------------------------------*/
void calculate_interaction_force_without_noise(soft_particles *particles,
                                               linked_cell_list *celllist,
                                               polymer_blend *blend,
                                               box_geometry *geometry)
{
    clear_interaction_force(particles);

    construct_linked_cell_list(celllist,particles,geometry);

    accumulate_site_interaction_force(particles,celllist,blend,geometry);
    accumulate_tethering_interaction_force(particles,blend,geometry);
}

/*---------------------------------------------------------------------------*/
void calculate_stress_tensor(soft_particles *particles,
                             linked_cell_list *celllist,
                             polymer_blend *blend,
                             box_geometry *geometry)
{
    clear_stress_tensor(particles);

    construct_linked_cell_list(celllist,particles,geometry);

    accumulate_site_interaction_stress(particles,celllist,blend,geometry);
    accumulate_tethering_interaction_stress(particles,blend,geometry);
}

/*---------------------------------------------------------------------------*/
EXPAND_FUNCTIONS(`dnl
void generate_random_force@D@(soft_particles@D@ *particles)
{
    int i,j;

    for(i = 0;i < particles->n;i++)
    {
        for(j = 0;j < particles->m[i];j++)
        {
            VECTOR@D@(particles->w[i][j][@@] = nrand();)
        }
    }
}

/*---------------------------------------------------------------------------*/
void clear_interaction_force@D@(soft_particles@D@ *particles)
{
    int i,j;

    for(i = 0;i < particles->n;i++)
    {
        for(j = 0;j < particles->m[i];j++)
        {
            VECTOR@D@(particles->f[i][j][@@] = 0.0;)

            VECTOR@D@(particles->f_star[i][j][@@] = 0.0;)
        }
    }
}

/*---------------------------------------------------------------------------*/
void clear_stress_tensor@D@(soft_particles@D@ *particles)
{
    int i,j;

    for(i = 0;i < particles->n;i++)
    {
        for(j = 0;j < particles->m[i];j++)
        {
            SYMMETRIC_TENSOR@D@(particles->sigma[i][j][@@] = 0.0;)
        }
    }
}'`')

dnl ---------------------------------------------------------------------------
dnl  GENERATE_FORCE_CALCULATION
dnl    generate functions to calculate the force or the stress tensor
dnl  $1 the function mode (force or stress).
dnl  $2 the type of the target argument (vector@D@ or symmetric_tensor@D@).
dnl  $3 the name of the target argument (f or sigma).
dnl ---------------------------------------------------------------------------
define(`GENERATE_FORCE_CALCULATION',`dnl
EXPAND_FUNCTIONS(`dnl
void accumulate_tethering_interaction_$1@D@(vector@D@ **r,$2 **$3,
                                               polymer_blend *blend,
                                               box_geometry *geometry)
{
    int i,j,k;
    vector@D@ dr;

    if(geometry->delta_l == 0.0)
    {
            /*  no box deformation  */
        for(i = 0;i < blend->n;i++)
        {
            if(blend->next[i] < 0)
                continue;

            k = blend->next[i];

                /*  calculate the tethering force.  */
            for(j = 0;j < blend->m[i];j++)
            {
                VECTOR@D@(dr[@@] = r[i][j][@@] - r[k][j][@@];)

                VECTOR@D@(if(dr[@@] >= 0.5 * geometry->l[@@])
                          {
                              dr[@@] -= geometry->l[@@];
                          }
                          else if(dr[@@] < - 0.5 * geometry->l[@@])
                          {
                              dr[@@] += geometry->l[@@];
                          })

                ifelse($1,`force',
                VECTOR@D@(f[i][j][@@] -= dr[@@];
                          f[k][j][@@] += dr[@@];),
                SYMMETRIC_TENSOR@D@(sigma[i][j][@@] += dr[@1@] * dr[@2@];
                                    sigma[k][j][@@] += dr[@1@] * dr[@2@];))
            }
        }
    }
    else
    {
            /*  box deformation  */
        for(i = 0;i < blend->n;i++)
        {
            if(blend->next[i] < 0)
                continue;

            k = blend->next[i];

                /*  calculate the tethering force.  */
            for(j = 0;j < blend->m[i];j++)
            {
                VECTOR@D@(dr[@@] = r[i][j][@@] - r[k][j][@@];)

                if(dr[1] >= 0.5 * geometry->l[1])
                {
                    dr[1] -= geometry->l[1];
                    dr[0] -= geometry->delta_l;
                }
                else if(dr[1] < - 0.5 * geometry->l[1])
                {
                    dr[1] += geometry->l[1];
                    dr[0] += geometry->delta_l;
                }
                while(dr[0] >= 0.5 * geometry->l[0])
                {
                    dr[0] -= geometry->l[0];
                }
                while(dr[0] < - 0.5 * geometry->l[0])
                {
                    dr[0] += geometry->l[0];
                }
                ifelse(@D@,3,
                       if(dr[2] >= 0.5 * geometry->l[2])
                       {
                           dr[2] -= geometry->l[2];
                       }
                       else if(dr[2] < - 0.5 * geometry->l[2])
                       {
                           dr[2] += geometry->l[2];
                       })

                ifelse($1,`force',
                VECTOR@D@(f[i][j][@@] -= dr[@@];
                          f[k][j][@@] += dr[@@];),
                SYMMETRIC_TENSOR@D@(sigma[i][j][@@] += dr[@1@] * dr[@2@];
                                    sigma[k][j][@@] += dr[@1@] * dr[@2@];))
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_site_interaction_$1@D@(vector@D@ **r,$2 **$3,
                                          linked_cell_list *celllist,
                                          polymer_blend *blend,
                                          box_geometry *geometry)
{
    ifelse(@D@,2,
           `int i1,j1,i2,j2,i2_i,j2_i;'`',
           @D@,3,
           `int i1,j1,k1,i2,j2,k2,i2_i,j2_i,k2_i;'`')
    int cell_index1,cell_index2;
    vector@D@ c;
    int i2_min,i2_max;

    if(geometry->delta_l == 0.0)
    {
            /*  no box deformation  */
        ifelse(@D@,3,
               for(k1 = 0;k1 < celllist->nz;k1++)
               {)
        for(j1 = 0;j1 < celllist->ny;j1++)
        {
            for(i1 = 0;i1 < celllist->nx;i1++)
            {
                ifelse(@D@,2,
                       cell_index1 = j1 * celllist->nx + i1;,
                       @D@,3,
                       cell_index1 = (k1 * celllist->ny + j1) * celllist->nx + i1;)

                ifelse(@D@,3,
                       for(k2 = k1 - 1;k2 <= k1 + 1;k2++)
                       {
                           k2_i = INDEX(k2,celllist->nz);)
                for(j2 = j1 - 1;j2 <= j1 + 1;j2++)
                {
                    j2_i = INDEX(j2,celllist->ny);

                    for(i2 = i1 - 1;i2 <= i1 + 1;i2++)
                    {
                        i2_i = INDEX(i2,celllist->nx);

                        ifelse(@D@,2,
                               cell_index2 = j2_i * celllist->nx + i2_i;,
                               @D@,3,
                               cell_index2 = (k2_i * celllist->ny + j2_i) * celllist->nx + i2_i;)

                        c[0] = LATTICE_VECTOR(i2,celllist->nx,geometry->l[0]);
                        c[1] = LATTICE_VECTOR(j2,celllist->ny,geometry->l[1]);
                        ifelse(@D@,3,
                               c[2] = LATTICE_VECTOR(k2,celllist->nz,geometry->l[2]);)


                        if(cell_index1 == cell_index2)
                        {
                            accumulate_site_interaction_$1@D@_in_same_cell(
                                r,$3,celllist,blend,cell_index1,&c);
                        }
                        else if(cell_index1 < cell_index2)
                        {
                            accumulate_site_interaction_$1@D@_in_different_cells(
                                r,$3,celllist,blend,cell_index1,cell_index2,&c);
                        }
                    }
                }
                ifelse(@D@,3,
                       })

            }
        }
        ifelse(@D@,3,
               })
    }
    else
    {
            /*  box deformation  */
        ifelse(@D@,3,
               for(k1 = 0;k1 < celllist->nz;k1++)
               {)

            /*  boundary cells (j = 0, ny - 1)  */
        for(j1 = 0;j1 < celllist->ny;j1 += (celllist->ny - 1))
        {
            for(i1 = 0;i1 < celllist->nx;i1++)
            {
                ifelse(@D@,2,
                       cell_index1 = j1 * celllist->nx + i1;,
                       @D@,3,
                       cell_index1 = (k1 * celllist->ny + j1) * celllist->nx + i1;)

                ifelse(@D@,3,
                       for(k2 = k1 - 1;k2 <= k1 + 1;k2++)
                       {
                           k2_i = INDEX(k2,celllist->nz);)
                for(j2 = j1 - 1;j2 <= j1 + 1;j2++)
                {
                    j2_i = INDEX(j2,celllist->ny);

                    if(IS_LOWER_BOUNDARY(j1,j2,celllist->ny))
                    {
                        i2_min = i1 + (int)floor(-1.0 + (geometry->delta_l * celllist->nx) / geometry->l[0]);
                        i2_max = i1 + (int)ceil(1.0 + (geometry->delta_l * celllist->nx) / geometry->l[0]);
                    }
                    else if(IS_UPPER_BOUNDARY(j1,j2,celllist->ny))
                    {
                        i2_min = i1 + (int)floor(-1.0 - (geometry->delta_l * celllist->nx) / geometry->l[0]);
                        i2_max = i1 + (int)ceil(1.0 - (geometry->delta_l * celllist->nx) / geometry->l[0]);
                    }
                    else
                    {
                        i2_min = i1 - 1;
                        i2_max = i1 + 1;
                    }

                    for(i2 = i2_min;i2 <= i2_max;i2++)
                    {
                        i2_i = i2;
                        c[0] = 0.0;

                        while(i2_i < 0)
                        {
                            i2_i += celllist->nx;
                            c[0] += geometry->l[0];
                        }
                        while(i2_i >= celllist->nx)
                        {
                            i2_i -= celllist->nx;
                            c[0] -= geometry->l[0];
                        }

                        if(IS_LOWER_BOUNDARY(j1,j2,celllist->ny))
                        {
                            c[0] += geometry->delta_l;
                        }
                        else if(IS_UPPER_BOUNDARY(j1,j2,celllist->ny))
                        {
                            c[0] -= geometry->delta_l;
                        }


                        ifelse(@D@,2,
                               cell_index2 = j2_i * celllist->nx + i2_i;,
                               @D@,3,
                               cell_index2 = (k2_i * celllist->ny + j2_i) * celllist->nx + i2_i;)

                        c[1] = LATTICE_VECTOR(j2,celllist->ny,geometry->l[1]);
                        ifelse(@D@,3,
                               c[2] = LATTICE_VECTOR(k2,celllist->nz,geometry->l[2]);)


                        if(cell_index1 == cell_index2)
                        {
                            accumulate_site_interaction_$1@D@_in_same_cell(
                                r,$3,celllist,blend,cell_index1,&c);
                        }
                        else if(cell_index1 < cell_index2)
                        {
                            accumulate_site_interaction_$1@D@_in_different_cells(
                                r,$3,celllist,blend,cell_index1,cell_index2,&c);
                        }
                    }
                }
                ifelse(@D@,3,
                       })
            }
        }


            /*  non-boundary cells  */
        for(j1 = 1;j1 < celllist->ny - 1;j1++)
        {
            for(i1 = 0;i1 < celllist->nx;i1++)
            {
                ifelse(@D@,2,
                       cell_index1 = j1 * celllist->nx + i1;,
                       @D@,3,
                       cell_index1 = (k1 * celllist->ny + j1) * celllist->nx + i1;)

                ifelse(@D@,3,
                       for(k2 = k1 - 1;k2 <= k1 + 1;k2++)
                       {
                           k2_i = INDEX(k2,celllist->nz);)
                for(j2 = j1 - 1;j2 <= j1 + 1;j2++)
                {
                    j2_i = INDEX(j2,celllist->ny);

                    for(i2 = i1 - 1;i2 <= i1 + 1;i2++)
                    {
                        i2_i = INDEX(i2,celllist->nx);

                        ifelse(@D@,2,
                               cell_index2 = j2_i * celllist->nx + i2_i;,
                               @D@,3,
                               cell_index2 = (k2_i * celllist->ny + j2_i) * celllist->nx + i2_i;)

                        c[0] = LATTICE_VECTOR(i2,celllist->nx,geometry->l[0]);
                        c[1] = LATTICE_VECTOR(j2,celllist->ny,geometry->l[1]);
                        ifelse(@D@,3,
                               c[2] = LATTICE_VECTOR(k2,celllist->nz,geometry->l[2]);)


                        if(cell_index1 == cell_index2)
                        {
                            accumulate_site_interaction_$1@D@_in_same_cell(
                                r,$3,celllist,blend,cell_index1,&c);
                        }
                        else if(cell_index1 < cell_index2)
                        {
                            accumulate_site_interaction_$1@D@_in_different_cells(
                                r,$3,celllist,blend,cell_index1,cell_index2,&c);
                        }
                    }
                }
                ifelse(@D@,3,
                       })

            }
        }
        ifelse(@D@,3,
               })
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_site_interaction_$1@D@_in_different_cells(
    vector@D@ **r,$2 **$3,
    linked_cell_list *celllist,polymer_blend *blend,
    int k1,int k2,vector@D@ *c)
{
    int i1,j1,i2,j2;
    double dr2;
    vector@D@ dr;
    double epsilon;

    for(i1 = 0;i1 < blend->n;i1++)
    {
        for(i2 = 0;i2 < blend->n;i2++)
        {
            ifelse(@D@,2,
                   epsilon = 1.0 / (M_PI) * blend->lambda0 * blend->epsilon[i1][i2];,
                   @D@,3,
                   epsilon = 1.0 / (M_PI * M_SQRTPI) * blend->epsilon[i1][i2];)

            j1 = celllist->head[i1][k1];
            while(!(j1 < 0))
            {
                j2 = celllist->head[i2][k2];
                while(!(j2 < 0))
                {
                    VECTOR@D@(dr[@@] = r[i1][j1][@@] - r[i2][j2][@@] + (*c)[@@];)

                    ifelse(@D@,2,
                           dr2 = SQR(dr[0]) + SQR(dr[1]);,
                           @D@,3,
                           dr2 = SQR(dr[0]) + SQR(dr[1]) + SQR(dr[2]);)

                    ifelse($1,`force',
                    VECTOR@D@(f[i1][j1][@@] += epsilon * exp(-dr2) * 2.0 * dr[@@];
                              f[i2][j2][@@] -= epsilon * exp(-dr2) * 2.0 * dr[@@];),
                    SYMMETRIC_TENSOR@D@(sigma[i1][j1][@@] -= epsilon * exp(-dr2) * 2.0 * dr[@1@] * dr[@2@];
                                        sigma[i2][j2][@@] -= epsilon * exp(-dr2) * 2.0 * dr[@1@] * dr[@2@];))

                    j2 = celllist->list[i2][j2];
                }

                j1 = celllist->list[i1][j1];
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void accumulate_site_interaction_$1@D@_in_same_cell(
    vector@D@ **r,$2 **$3,
    linked_cell_list *celllist,polymer_blend *blend,
    int k,vector@D@ *c)
{
    int i1,j1,i2,j2;
    double dr2;
    vector@D@ dr;
    double epsilon;

    for(i1 = 0;i1 < blend->n;i1++)
    {
            /*  different species  */
        for(i2 = 0;i2 < i1;i2++)
        {
            ifelse(@D@,2,
                   epsilon = 1.0 / (M_PI) * blend->lambda0 * blend->epsilon[i1][i2];,
                   @D@,3,
                   epsilon = 1.0 / (M_PI * M_SQRTPI) * blend->epsilon[i1][i2];)

            j1 = celllist->head[i1][k];
            while(!(j1 < 0))
            {
                j2 = celllist->head[i2][k];
                while(!(j2 < 0))
                {
                    VECTOR@D@(dr[@@] = r[i1][j1][@@] - r[i2][j2][@@] + (*c)[@@];)

                    ifelse(@D@,2,
                           dr2 = SQR(dr[0]) + SQR(dr[1]);,
                           @D@,3,
                           dr2 = SQR(dr[0]) + SQR(dr[1]) + SQR(dr[2]);)

                    ifelse($1,`force',
                    VECTOR@D@(f[i1][j1][@@] += epsilon * exp(-dr2) * 2.0 * dr[@@];
                              f[i2][j2][@@] -= epsilon * exp(-dr2) * 2.0 * dr[@@];),
                    SYMMETRIC_TENSOR@D@(sigma[i1][j1][@@] -= epsilon * exp(-dr2) * 2.0 * dr[@1@] * dr[@2@];
                                        sigma[i2][j2][@@] -= epsilon * exp(-dr2) * 2.0 * dr[@1@] * dr[@2@];))


                    j2 = celllist->list[i2][j2];
                }

                j1 = celllist->list[i1][j1];
            }
        }

            /*  same species  */
        ifelse(@D@,2,
               epsilon = 1.0 / (M_PI) * blend->lambda0 * blend->epsilon[i1][i1];,
               @D@,3,
               epsilon = 1.0 / (M_PI * M_SQRTPI) * blend->epsilon[i1][i1];)

        j1 = celllist->head[i1][k];
        while(!(j1 < 0))
        {
            j2 = celllist->head[i1][k];
            while(!(j2 < 0))
            {
                if(j1 < j2)
                {
                    VECTOR@D@(dr[@@] = r[i1][j1][@@] - r[i1][j2][@@] + (*c)[@@];)

                    ifelse(@D@,2,
                           dr2 = SQR(dr[0]) + SQR(dr[1]);,
                           @D@,3,
                           dr2 = SQR(dr[0]) + SQR(dr[1]) + SQR(dr[2]);)

                    ifelse($1,`force',
                    VECTOR@D@(f[i1][j1][@@] += epsilon * exp(-dr2) * 2.0 * dr[@@];
                              f[i1][j2][@@] -= epsilon * exp(-dr2) * 2.0 * dr[@@];),
                    SYMMETRIC_TENSOR@D@(sigma[i1][j1][@@] -= epsilon * exp(-dr2) * 2.0 * dr[@1@] * dr[@2@];
                                        sigma[i1][j2][@@] -= epsilon * exp(-dr2) * 2.0 * dr[@1@] * dr[@2@];))

                }

                j2 = celllist->list[i1][j2];
            }

            j1 = celllist->list[i1][j1];
        }
    }
}'`')'`')dnl
dnl
dnl ---------------------------------------------------------------------------
dnl
/*---------------------------------------------------------------------------*/
GENERATE_FORCE_CALCULATION(force,vector@D@,f)

/*---------------------------------------------------------------------------*/
GENERATE_FORCE_CALCULATION(stress,symmetric_tensor@D@,sigma)

/*---------------------------------------------------------------------------*/
